;****************************************************************************
; BCD addition algorithm courtesy Douglas W. Jones at University Of Iowa
; Details available at http://www.cs.uiowa.edu/~jones/bcd/bcd.html#packed
; This is my implementation of his posted C algorithm.
;****************************************************************************

; Quick and dirty implementation ... uses globals and preserves nothing!

; Called with:
;   esi -> valid BCD source operands
;   edi -> valid BCD destination operands
;   wksize = number of dwords used in both source and destination
;   ddsize = maximum number of dwords available for each
; Returns with:
;   destination = source + destination
;   wksize incremented if another dword was needed to store final carry
;   carry clear if no errors occured
;   carry flag set on error
;   all registers destroyed


OPTION PROLOGUE:NONE 
OPTION EPILOGUE:NONE 

        align   16

TestA   proc C

        mov     eax,wksize              ; Number of dwords currently used
        mov     ecx,eax
        shl     eax,2
        neg     ecx                     ; Loop counts up from minus wksize
        add     esi,eax                 ; Point 1 dword past most significant
        add     edi,eax                 ;   source and destination
nocry:  mov     eax,[esi+4*ecx]         ; Start with no carry-in for least
        mov     ebx,[edi+4*ecx]         ;  significant source and destination
        mov     edx,eax                 ;
        add     eax,66666666h           ; eax = source + excess_6
        jc      foul                    ; No carry for 6's + 9's, only F's
        add     eax,ebx                 ; eax = dst+src+excess_6
        jc      cry1                    ; Do carry-out if set
nocry1: xor     edx,ebx
        xor     edx,eax
        not     edx
        and     edx,11111110h           ; Leave a 1 for every BCD digit
        mov     ebx,edx                 ;  that did not produce a carry-out
        shr     edx,2
        shr     ebx,3
        or      edx,ebx
        or      edx,60000000h           ; Leave a 6 for each digit that did
        sub     eax,edx                 ;  not carry and remove excess_6
        mov     [edi+4*ecx],eax         ; Store valid result in destination
        add     ecx,1                   ; Count up to 0 and do each dword
        jnz     nocry                   ; Remember, no carry-in for the next
        jmp short ok

cryx:   add     eax,ebx                 ; Source was 99999999 + carry-in
        jmp short cry1

cry:    mov     eax,[esi+4*ecx]
        mov     ebx,[edi+4*ecx]
        mov     edx,eax
        add     eax,66666667h           ; Excess_6 + carry-in
        jc      cryx
        add     eax,ebx
        jnc     nocry1                  ; No carry-out or carry-in for next
cry1:   xor     edx,ebx
        xor     edx,eax
        not     edx
        and     edx,11111110h           ; Same as above except all digits
        mov     ebx,edx                 ;  can be handled with mask because
        shr     edx,2                   ;  the most significant digit DID
        shr     ebx,3                   ;  produce a carry and doesn't need
        or      edx,ebx                 ;  to be adjusted to remove excess_6
        sub     eax,edx
        mov     [edi+4*ecx],eax         ; Store valid result in destination
        add     ecx,1
        jnz     cry                     ; Do carry-in for next

        mov     eax,wksize              ; Send carry to the next dword
        cmp     eax,ddsize
        jae     foul                    ; Oops! Not enough dwords available
        add     wksize,1                ; Bump number of dwords being used
        mov dword ptr [esi],0           ; Clear next source
        mov dword ptr [edi],1           ; Store the carry and return

ok:     clc
        ret

foul:   stc
        ret

TestA   endp

OPTION PROLOGUE:PrologueDef
OPTION EPILOGUE:EpilogueDef 
